import torch
from torch_geometric.nn import MessagePassing

class OPE(MessagePassing):
    """
    Computes:
    \|(\nabla^{+}_{w}f(v_{i})\|_{p} = \left [ \sum_{v_{j}\sim v_{i}} w(v_{i}v_{j})^{\frac{p}{2}}((f(v_j) - f(v_i))^{+})^{p}  \right]^{\frac{1}{p}}
    """
    def __init__(self, graph, **kwargs):
        super(OPE, self).__init__(aggr='add') 
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr
        self.p = kwargs["p_val"]

    def forward(self, x):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        return torch.pow(torch.sqrt(self.edge_attr)  * torch.clamp(x_j-x_i, 0), self.p) 

    def update(self, aggr_out):
        return torch.pow(aggr_out, 1/self.p)



def info():
    about = """
    This subpackage implements operators on graphs to run a manually defined PDE on graph, Calculus on graphs.

    Ref:
    https://en.wikipedia.org/wiki/Calculus_on_finite_weighted_graphs
    """
    print(about)

__all__ = [
        'GradInfNorm',
        'GradMinusInfNorm',
        'GradMinusNorm',
        'GradNorm',
        'GradPlusInfNorm',
        'GradPlusNorm',
        'MeanCurv',
        'LapIso',
        'LapAniso',
        'info',

]



"""
Runs a predefined isotorpic diffusion PDE on graph
Implements the following iterative scheme:

\mathbf{x}^{n+1}_{i} = \frac{\sum_{j \in N(i)} w_{i,j}^{p/2}\|\mathbf{x}^{n}_{j}-\mathbf{x}^{n}_{i}\|_{2}^{p-2}\mathbf{x}^{n}_{j}}{\sum_{j \in N(i)} w_{i,j}^{p/2}\|\mathbf{x}^{n}_{j}-\mathbf{x}^{n}_{i}\|_{2}^{p-2}}

Ref:
(https://hal.archives-ouvertes.fr/hal-00163573/document)
"""
import torch
from torch_geometric.nn import MessagePassing
from tqdm import tqdm

class Numerator(MessagePassing):
    def __init__(self, graph, **kwargs):
        super(Numerator, self).__init__(aggr='add') 
        self.ep = kwargs["epsilon"]
        self.p = kwargs["p_val"]
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self, x):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)
              
    def message(self, x_i, x_j):
        a = x_j - x_i
        b = torch.norm(a, dim=1)
        b = b * b
        tmp1 =torch.pow((b+self.ep), (self.p-2)/2)
        tmp2 = torch.stack((tmp1,tmp1,tmp1),dim=1)
        tmp = torch.pow(torch.sqrt(self.edge_attr), self.p) * tmp2 * x_j
        return tmp

    def update(self, aggr_out):
        return  aggr_out

class Denominator(MessagePassing):
    def __init__(self,graph, **kwargs):
        super(Denominator, self).__init__(aggr='add') 
        self.ep = kwargs["epsilon"]
        self.p = kwargs["p_val"]
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self, x):
        edge_index = self.edge_index
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        a = x_j - x_i
        b = torch.norm(a, dim=1)
        b = b * b
        tmp1 = torch.pow((b+self.ep),(self.p-2)/2)
        tmp2 = torch.stack((tmp1,tmp1,tmp1),dim=1)
        tmp = torch.pow(torch.sqrt(self.edge_attr), self.p) * tmp2 
        return tmp

    def update(self, aggr_out):
        return  aggr_out


class PDE(torch.nn.Module):
    def __init__(self, graph, **kwargs):
        super(PDE, self).__init__()
        self.num = Numerator(graph, **kwargs)
        self.den = Denominator(graph, **kwargs)

    def forward(self,signal,itr):
        for i in tqdm(range(itr)):
            u =  self.num(signal)
            d =  self.den(signal)
            signal = u / d
        return signal



"""
Runs a predefined anisotropic diffusion PDE on graph:
Implements the following iterative scheme:

\mathbf{x}^{n+1}_{i} = \frac{\lambda \mathbf{x}^{0}_{i} + \sum_{j \in N(i)} w_{i,j}^{p/2}|\mathbf{x}^{n}_{j}-\mathbf{x}^{n}_{i}|^{p-2}\mathbf{x}^{n}_{j}}{\lambda+ \sum_{j \in N(i)} w_{i,j}^{p/2}|\mathbf{x}^{n}_{j}-\mathbf{x}^{n}_{i}|^{p-2}}

Ref:
https://hal.archives-ouvertes.fr/hal-00163573/document
"""

import torch
from torch_geometric.nn import MessagePassing
from tqdm import tqdm

class Numerator(MessagePassing):
    def __init__(self, graph, **kwargs):
        super(Numerator, self).__init__(aggr='add') #  "add" aggregation.
        self.ep = kwargs["epsilon"]
        self.lamb = kwargs["lamb"]
        self.p = kwargs["p_val"]
        self.x0 = kwargs["x0"]
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self, x):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        tmp = torch.pow(torch.sqrt(self.edge_attr), self.p) * torch.pow(torch.abs((x_j - x_i) + self.ep), self.p-2) * x_j 
        return tmp

    def update(self, aggr_out):
        return self.x0 * self.lamb  + aggr_out

class Denominator(MessagePassing):
    def __init__(self, graph, **kwargs):
        super(Denominator, self).__init__(aggr='add') #  "add" aggregation.
        self.ep = kwargs["epsilon"]
        self.lamb = kwargs["lamb"]
        self.p = kwargs["p_val"]
        self.x0 = kwargs["x0"]
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self, x):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        tmp =  torch.pow(torch.sqrt(self.edge_attr), self.p) * torch.pow(torch.abs((x_j - x_i) + self.ep),self.p-2)
        return tmp

    def update(self, aggr_out):
        return self.lamb + aggr_out


class PDE(torch.nn.Module):
    def __init__(self, graph, **kwargs):
        super(PDE, self).__init__()
        self.num = Numerator(graph, **kwargs)
        self.den = Denominator(graph, **kwargs)

    def forward(self,signal,itr):
        for i in tqdm(range(itr)):
            u =  self.num(signal)
            d =  self.den(signal)
            signal = u / d
        return signal



def info():
    about = """
    This subpackage implements predefined PDES on graphs.

    Ref:
    https://elmoatazbill.users.greyc.fr/pub.html
    """
    print(about)

__all__ = [
        'pdeanisodiff',
        'pdeisodiff',
        'pdeinflapdiff',
        'pdemorphdilation',
        'pdemorpherosion',
        'info',
]



import torch
from torch_geometric.nn import MessagePassing
from ..operators import GradNorm

class OPE(MessagePassing): 
    """
    Computes:
    \sum_{j\in N(i)} w_{i,j}\ (\|\nabla_{w}\mathbf{x}_{i}\|_{2}^{p-2} + \|\nabla_{w}\mathbf{x}_{j}\|_{2}^{p-2})\  (\mathbf{x}_{j} - \mathbf{x}_{i})
    """
    def __init__(self, graph, **kwargs):
        super(OPE, self).__init__(aggr='add') 
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr
        self.ope = GradNorm.OPE(graph, **kwargs)
        self.p = kwargs["p_val"]
        
    def forward(self, signal):
        new_sig = self.ope(signal)
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x, gradSig = new_sig) 

    def message(self, x_i, x_j, gradSig_i, gradSig_j):
        tmp = 0.5 * self.edge_attr * (torch.pow(gradSig_i, self.p-2) + torch.pow(gradSig_i, self.p-2)) * (x_j - x_i)
        return tmp



__version__ = "1.0.4"

def info():
    about = """
    Torch_pdegraph is a proof of concept that how one can solve PDEs on graphs using torch-geometric's Message Passing classes.
    """
    print(about)

__all__= [
        '__version__',
        'info',
        ]



import torch
from torch_geometric.nn import MessagePassing

class OPE(MessagePassing): 
    """
    Computes:
    \sum_{j\in N(i)} w_{i,j}^{\frac{p}{2}}\  |\mathbf{x}_{j} - \mathbf{x}_{i} |^{p-2} \  (\mathbf{x}_{j} - \mathbf{x}_{i})

    """
    def __init__(self, graph, **kwargs):
        super(OPE, self).__init__(aggr='add') 
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr
        self.p = kwargs["p_val"]
        
    def forward(self, signal):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        tmp = torch.pow(torch.sqrt(self.edge_attr), self.p) * torch.pow(torch.abs(x_j - x_i), self.p - 2) * (x_j - x_i)
        return tmp



"""
Runs a predefined infinitylaplacian diffusion PDE on graph:

Solves the following problem:
  \left\{\begin{matrix}
      \frac{\partial f(u, t)}{\partial t}
        = \Delta_{\omega, \infty}(f)(u)
    \\
      f(u, 0) = f^{0}(u)
    \end{matrix}\right.

Implements the following iterative scheme:
    f^{n+1}(u) = f^{n}(u) + 0.5 * ( \left\|\nabla_{\omega}^{+}(f)(u) \right\|_{\infty} - \left\|\nabla_{\omega}^{-}(f)(u) \right\|_{\infty})
    Or
    f^{n+1}(u) = f^{n}(u) + 0.5 * (nld(signal) - nle(signal)

Ref:
Abderrahim Elmoataz, Xavier Desquesnes, Zakaria Lakhdari,
and Olivier Lezoray. On the infinity laplacian equation on
graph with applications to image and manifolds processing.
In International Conference on Approximation Methods and
Numerical Modelling in Environment and Natural Resources,
pages 7–pages, 2011.  
"""
import torch
from ..operators import GradMinusInfNorm
from ..operators import GradPlusInfNorm
from tqdm import tqdm

class PDE(torch.nn.Module):
    def __init__(self, graph, **kwargs):
        super(PDE, self).__init__():
        self.ope_m = GradMinusInfNorm.OPE(graph, **kwargs)
        self.ope_p = GradPlusInfNorm.OPE(graph, **kwargs)

    def forward(self,signal,itr):
        for i in tqdm(range(itr)):
            signal = signal +  0.5 * (self.ope_p(signal) - self.ope_m(signal))
        return signal



import torch
from torch_geometric.nn import MessagePassing

class OPE(MessagePassing):
    """
    Computes: 
    \|(\nabla_{w}^{-}f(v_{i})\|_{\infty} = \max_{v_{j}\sim v_{i}}\sqrt{w(v_{i},v_{j})}((f(v_j) - f(v_i))^{-} 
    """
    def __init__(self, graph, **kwargs):
        super(OPE, self).__init__(aggr='max') 
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self, x):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        tmp = torch.sqrt(self.edge_attr)  * torch.clamp(x_i-x_j, 0.00)
        return tmp 

    def update(self, aggr_out):
        
        return aggr_out 



import pathlib
from setuptools import setup, find_packages

HERE =  pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

install_requires = [
    'torch',
    'numpy',
    'mnist',
    'tqdm',
    'scikit-build',
    'scipy',
    'scikit-learn',
    'scikit-image',
    'matplotlib',
    'open3d-python',
]

setup(
    name="torch_pdegraph",
    version="1.0.4",
    description="Running paritial difference equations (PDEs) on graphs",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/aGIToz/Pytorch_pdegraph",
    python_requires='>=3.6',
    install_requires=install_requires,
    packages=find_packages(),
    author="Amitoz AZAD",
    author_email="amitoz.sudo@gmail.com",
    license="MIT",
    include_package_data=True,
    )



"""
Runs a predefined morphological dilation on graphs.

Solves the following problem:
    \left\{\begin{matrix}
      \frac{\partial f(u, t)}{\partial t}
        = \left\|\nabla_{\omega}^{+}(f)(u) \right\|_{\infty}
    \\
      f(u, 0) = f^{0}(u)
    \end{matrix}\right.

Implements the following iterative scheme:
    f^{n+1}(u) = f^{n}(u) + \left\|\nabla_{\omega}^{+}(f)(u) \right\|_{\infty}

Ref:
  V.-T. Ta, A. Elmoataz et O. Lézoray :<br>
  Nonlocal pdes-based morphology on weighted<br>
  graphs for image and data processing. IEEE<br>
  Trans. Image Process., 20(6):1504–1516, 2011.
"""

import torch
from ..operators import GradPlusInfNorm
from tqdm import tqdm

class PDE(torch.nn.Module):
    def __init__(self, graph, **kwargs):
        super(PDE, self).__init__()
        self.ope = GradPlusInfNorm.OPE(graph, **kwargs)

    def forward(self,signal,itr):
        for i in tqdm(range(itr)):
            signal = signal + self.ope(signal)
        return signal



import torch
from torch_geometric.nn import MessagePassing

class Numerator(MessagePassing):
    def __init__(self, graph, **kwargs):
        super(Numerator, self).__init__(aggr='add') 
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self, x):
        return self.propagate(edge_index=self.edge_index, size=(x.size(0), x.size(0)), x=x)

    def message(self, x_i, x_j):
        return torch.sqrt(self.edge_attr)  * torch.sign(x_j-x_i) 

class Denominator(MessagePassing):
    def __init__(self, graph, **kwargs):
        super(Denominator, self).__init__(aggr='add') 
        self.edge_index = graph.edge_index
        self.edge_attr = graph.edge_attr

    def forward(self):
        return self.propagate(edge_index = self.edge_index)

    def message(self):
        return torch.sqrt(self.edge_attr) 

class OPE(torch.nn.Module):
    """
    Computes: 
    \kappa_{w}(u,f) = \frac{\sum_{v\sim u}\sqrt{w(u,v)}sign(f(v)-f(u))}{\sum_{v\sim u}\sqrt{w(u,v)}}
    """
    def __init__(self, graph, **kwargs):
        super(OPE, self).__init__() 
        self.num = Numerator(graph, **kwargs)
        self.den = Denominator(graph, **kwargs)

    def forward(self, signal):
        signal = self.num(signal)/self.den()
        return signal



"""
Azad Amitoz, 2020-08-17
This module allows some basic operations on images and poinclouds. 
Helps creating initial generalized-distances and level-sets.
"""

import numpy as np
from matplotlib import pyplot as plt
import open3d as o3d

def toFmat(img):
    """
    Converts to your image to feature matrix of size (N by c), N is the 
    number of pixels and c is number of channels

    param img: A gray-scale or multi-channel image.
    return fmat: A (N,c) size feature matrix.
    """
    try:
        l, w, c = img.shape
        fmat = np.ones((l*w,c)) 
        for i in range(c): fmat[:,i] = np.reshape(img[:,:,i],(l*w,))
    except ValueError:
        try:
            l, w = img.shape
            fmat = np.ones((l*w,1)) 
            fmat = np.reshape(img,(l*w,1))
        except ValueError:
            print("Image should be at least a 2D array.")
    return fmat

def toPmat(shape):
    """
    Returns the meshgrid of shape (N, M) as position vector of shape (N*M,2)

    param shape: A tuple, shape of an image say of size  (N,M)
    return pmat: postion vector of shape of size (N*M, 2) 
    """
    x = np.arange(0,shape[1],1)
    y = np.arange(shape[0],0,-1)
    meshx, meshy = np.meshgrid(x,y)
    x = np.reshape(meshx,(shape[0]*shape[1],1)) 
    y = np.reshape(meshy,(shape[0]*shape[1],1)) 
    pmat = np.concatenate((x,y),axis=1) / max(shape[0],shape[1])
    return pmat

def getWgStats(wg):
    return f"The max is {np.max(wg)}; The min is {np.min(wg)}; The std is {np.std(wg)};  The median is {np.median(wg)}; The mean is {np.mean(wg)}"

def addNoise(mu, img):
    """
    Adds some white guassian noise

    param mu: What percentage of noise to be added relative the max signal in image
    param img:  Input image
    return signal: The noisy image
    """
    mu = (mu * np.max(img))/100 # TAKE VALUE ACC TO THE MAX VALUE IN IMAGE
    noise = np.random.normal(0, mu, img.shape)
    signal = noise + img
    return signal

def toImg(fmat, shape):
    """
    Converts the (N,c) feature matrix back to original image

    param fmat: The feature matrix of size (N,c)
    param shape: The tuple (N1, N2) assuming N1*N2 = N
    return img: The (N1,N2) shaped image 
    """
    img = np.zeros(shape)
    try:
        l, w, c = img.shape
        for i in range(c): img[:,:,i] = np.reshape(fmat[0:,i],(l,w))
    except ValueError:
        try:
            l, w = img.shape
            img = np.reshape(fmat,(l,w))
        except ValueError:
            print("Image should be at least a 2D array.")
    return img

def getPSNR(imgr, imgt):    
    """
    To get the PSNR

    param imgr: Reference image
    param imgt: Target image
    """
    mse = np.mean( (imgr - imgt) ** 2 )
    if mse == 0: return 100
    PIXEL_MAX = 255.0
    return 20 * math.log10(PIXEL_MAX / math.sqrt(mse))

def imgPatches(img, patch_shape=(3, 3)):
    """
    Generates the image patches.

    param img: The input image containing N-pixels.
    param patch_shape: The shape of the patches required, say (m,n)
    return patches: The patches feature vector in shape (N, m*n)
    """
    try:
        h, w, d = img.shape
    except ValueError:
        try:
            h, w = img.shape
            d = 0
        except ValueError:
            print("Image should be at least a 2D array.")
    try:
        r, c = patch_shape
    except ValueError:
        r = patch_shape[0]
        c = r
    pad_width = [(int((r - 0.5) / 2.), int((r + 0.5) / 2.)),
                 (int((c - 0.5) / 2.), int((c + 0.5) / 2.))]
    if d == 0:
        window_shape = (r, c)
        d = 1  # For the reshape in the return call
    else:
        pad_width += [(0, 0)]
        window_shape = (r, c, d)

    # Pad the image.
    img = np.pad(img, pad_width=pad_width, mode='symmetric')

    # Extract patches as node features.
    try:
        import skimage
    except Exception:
        raise ImportError('Cannot import skimage, which is needed to '
                          'extract patches. Try to install it with '
                          'pip (or conda) install scikit-image.')
    patches = skimage.util.view_as_windows(img, window_shape=window_shape)
    patches = patches.reshape((h * w, r * c * d))
    return patches

def dispImg(img): 
    """
    This does the min-max scaling the images to 0 and 1 and displays the image.

    param img: The input image
    return None: Just displays the image in the notebook.
    """
    try:
        h, w, d = img.shape
        img_tmp = (img - np.min(img)) / (np.max(img) - np.min(img))
        plt.matshow(img_tmp)
        plt.show(block=False)
    except ValueError:
        try:
            h, w = img.shape
            img_tmp = (img - np.min(img)) / (np.max(img) - np.min(img))
            plt.matshow(img_tmp, cmap = "gray")
            plt.show(block=False)
        except ValueError:
            print("Image should be at least a 2D array.")
    return None  

def downsamplepcd(**kwargs):
    """
    This does the downsampling of pcd

    param position: The (N,3) matrix
    param texture: The (N,3) matrix
    param voxel_size: The small number < 1, should be adjusted acc downsampling requirement
    returns: The downsampled position and texture
    """
    p = kwargs["position"]
    t = kwargs["texture"]
    voxel_size = kwargs["voxel_size"]
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(p)
    pcd.colors = o3d.utility.Vector3dVector(t)
    downpcd = pcd.voxel_down_sample(voxel_size)
    p = downpcd.points
    t = downpcd.colors
    return np.asarray(p), np.asarray(t)

def displayJSur(**kwargs):
    """
    Displays the pcd/mesh inside the Jupyter Notebook
    
    param position: The (N,3) matrix
    param texture: The (N,3) matrix
    """
    p = kwargs["position"]
    t = kwargs["texture"]
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(p)
    pcd.colors = o3d.utility.Vector3dVector(t)
    visualizer = o3d.JVisualizer()
    visualizer.add_geometry(pcd)
    visualizer.show()
    return None

def displaySur(**kwargs):
    """
    Displays the pcd outside the Jupyter Notebook

    param position: The (N,3) matrix
    param texture: The (N,3) matrix
    """
    p = kwargs["position"]
    t = kwargs["texture"]
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(p)
    pcd.colors = o3d.utility.Vector3dVector(t)
    o3d.visualization.draw_geometries([pcd]) 
    return None

def getPositionTexture(file_path):
    """
    Takes the input as pcd/mesh file and returns the vertices, texture, faces.
    """
    mesh = o3d.io.read_triangle_mesh(file_path)
    vertices = np.asarray(mesh.vertices)
    texture = np.asarray(mesh.vertex_colors) if len(np.asarray(mesh.vertex_colors)) != 0 else np.ones(vertices.shape)*0.5
    triangles = np.asarray(mesh.triangles)
    return (vertices, texture, triangles)


def dispMesh(**kwargs):
    """
    Displays the mesh outside the Jupyter environment
    """
    p = kwargs["position"]
    t = kwargs["texture"]
    f = kwargs["faces"]
    mesh = o3d.geometry.TriangleMesh()
    mesh.vertices = o3d.utility.Vector3dVector(p)
    mesh.vertex_colors = o3d.utility.Vector3dVector(t)
    mesh.triangles = o3d.utility.Vector3iVector(f)
    o3d.visualization.draw_geometries([mesh])
    return None


def getDistPcd(texture):
    """
    To get the intial distance on the pcd, used in eikonal equation

    param texture: This input texture should mostly be zero except where you annotate (that is put seed)

    return mask: Wherever you did the annotation, the mask is  0 at that place  and 1 elsewhere.
    """
    mask = (texture[:,:] != 0).astype("int")
    mask = np.sum(mask, axis=1)
    mask[mask[:] == 0] = -1
    mask[mask[:] > 0] = 0
    mask[mask[:] == -1] = 1
    return mask

def getDist(img):
    """
    To get the intial distance on the image, used in eikonal equation used for example in segmentation.

    param img: This input img should mostly be zero except where you annotate (that is put seed) 
    return mask: Wherever you did the annotation, the mask is  0 at that place  and 1 elsewhere. 
    """
    mask = (img[:,:,:] != 0).astype("int")
    mask = np.sum(mask, axis=2)
    mask[mask[:,:] == 0] = -1
    mask[mask[:,:] > 0] = 0
    mask[mask[:,:] == -1] = 1
    return mask

def genInitialSeeds(**kwargs):
    """
    Generates the intial level-set, used for example in problems of semi-supervised classification.

    param labels: The (N,1) vector N being the number of samples to be classified.
    param num_seeds: The number of labels (semi-supervised instances) you want to put. 
    return Front: It is the list of len C, C being the number of classes, each element is front corresponding to a class of shape (N,1),
    wherever you put the labels, the value is 1 and -1 elsewhere. (Values inside the level-set = 1 and outside = -1)
    """
    labels = kwargs["labels"]
    num_seeds = kwargs["num_seeds"]

    M = []
    for i in range(np.max(labels)+1):
        mask = (labels == i)
        dist = np.ones(len(mask))
        M.append((mask,dist))

    Front = []
    for mTuple in M:
        j = 0
        for i in range(len(mTuple[0])):
            if j < num_seeds:
                if mTuple[0][i] == True:
                    mTuple[1][i] = -1
                    j = j+1
            else: break
        Front.append(-1 * np.reshape(mTuple[1], (len(mTuple[1]),1)))
    del M # not required
    print("Success!")
    return Front
    
from collections import namedtuple
#A simple structure to group the edge_index and edge_attr to create a graph.
Graph = namedtuple('Graph', 'edge_index, edge_attr')
